#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

__global__ void unrolled_matmul_pool_kernel(
    const float* __restrict__ x,
    const float* __restrict__ weight,
    const float* __restrict__ bias,
    float* __restrict__ output,
    int B,           // batch_size
    int inF,         // number of input features
    int outF,        // number of output features
    int kernel_size,
    float scale_factor
) {
    const int tid = threadIdx.x;
    const int b = blockIdx.x;
    if (b >= B) return;

    extern __shared__ float smem[];
    float* linear_out = smem;                    // size: outF
    float* warp_results = &smem[outF];          // size: blockDim.x/32

    // Linear transformation with manual unrolling for the inner loop
    #pragma unroll 4
    for (int j = tid; j < outF; j += blockDim.x) {
        float sum = bias[j];
        const float* w_row = weight + j * inF;
        const float* x_row = x + b * inF;
        
        // Manual unroll of the inner product loop in chunks of 4
        int i = 0;
        for (; i + 3 < inF; i += 4) {
            sum = __fmaf_rn(x_row[i], w_row[i], sum);
            sum = __fmaf_rn(x_row[i+1], w_row[i+1], sum);
            sum = __fmaf_rn(x_row[i+2], w_row[i+2], sum);
            sum = __fmaf_rn(x_row[i+3], w_row[i+3], sum);
        }
        // Handle remaining elements
        for (; i < inF; i++) {
            sum = __fmaf_rn(x_row[i], w_row[i], sum);
        }
        linear_out[j] = sum;
    }
    __syncthreads();

    // Max pooling with manual unrolling for common kernel sizes
    int pooled_len = (outF >= kernel_size) ? ((outF - kernel_size) / kernel_size + 1) : 0;
    float thread_sum = 0.0f;

    #pragma unroll 4
    for (int seg = tid; seg < pooled_len; seg += blockDim.x) {
        const int start_idx = seg * kernel_size;
        float max_val = linear_out[start_idx];

        // Manually unrolled max computation based on common kernel sizes
        if (kernel_size == 2) {
            max_val = fmaxf(max_val, linear_out[start_idx + 1]);
        }
        else if (kernel_size == 4) {
            max_val = fmaxf(max_val, linear_out[start_idx + 1]);
            max_val = fmaxf(max_val, linear_out[start_idx + 2]);
            max_val = fmaxf(max_val, linear_out[start_idx + 3]);
        }
        else if (kernel_size == 8) {
            max_val = fmaxf(max_val, linear_out[start_idx + 1]);
            max_val = fmaxf(max_val, linear_out[start_idx + 2]);
            max_val = fmaxf(max_val, linear_out[start_idx + 3]);
            max_val = fmaxf(max_val, linear_out[start_idx + 4]);
            max_val = fmaxf(max_val, linear_out[start_idx + 5]);
            max_val = fmaxf(max_val, linear_out[start_idx + 6]);
            max_val = fmaxf(max_val, linear_out[start_idx + 7]);
        }
        else {
            #pragma unroll 4
            for (int k = 1; k < kernel_size; k++) {
                max_val = fmaxf(max_val, linear_out[start_idx + k]);
            }
        }
        thread_sum += max_val;
    }

    // Warp-level reduction using shuffle operations
    const unsigned int lane_id = tid % 32;
    const unsigned int warp_id = tid / 32;
    
    // Reduce within warp using explicitly unrolled shuffle operations
    #pragma unroll
    for (int offset = 16; offset > 0; offset >>= 1) {
        thread_sum += __shfl_down_sync(0xffffffff, thread_sum, offset);
    }

    // First thread in each warp writes to shared memory
    if (lane_id == 0) {
        warp_results[warp_id] = thread_sum;
    }
    __syncthreads();

    // Final reduction across warps by first warp
    if (tid < 32) {
        float warp_sum = (tid < (blockDim.x / 32)) ? warp_results[tid] : 0.0f;
        
        // Explicit unroll of final warp reduction
        #pragma unroll
        for (int offset = 16; offset > 0; offset >>= 1) {
            warp_sum += __shfl_down_sync(0xffffffff, warp_sum, offset);
        }

        if (tid == 0) {
            output[b] = warp_sum * scale_factor;
        }
    }
}

at::Tensor forward(
    at::Tensor x,
    int64_t kernel_size,
    double scale_factor,
    at::Tensor weight,
    at::Tensor bias
) {
    TORCH_CHECK(x.is_cuda(), "x must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(bias.is_cuda(), "bias must be a CUDA tensor");
    TORCH_CHECK(x.dim() == 2, "x must have shape (batch_size, in_features)");
    TORCH_CHECK(weight.dim() == 2, "weight must have shape (out_features, in_features)");
    TORCH_CHECK(bias.dim() == 1, "bias must have shape (out_features)");

    const auto B = x.size(0);
    const auto inF = x.size(1);
    const auto outF = weight.size(0);

    auto out = torch::empty({B}, x.options());

    const int threads = 256;
    const size_t smem_size = (outF + (threads/32)) * sizeof(float);

    unrolled_matmul_pool_kernel<<<B, threads, smem_size>>>(
        x.data_ptr<float>(),
        weight.data_ptr<float>(),
        bias.data_ptr<float>(),
        out.data_ptr<float>(),
        B,
        inF,
        outF,
        static_cast<int>(kernel_size),
        static_cast<float>(scale_factor)
    );

    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Unrolled CUDA implementation");
}